
/**************************************
     Simple SD/MMC audio recorder
 -------------------------------------
 Copyright 2011 David Guillen Fandos
 Use at your own risk!
 -------------------------------------
 Records audio from input line to the
 card. The card must be FAT16/32 and 
 a file called F1 must exist. The F1
 file will be filled with a WAV-header
 plus the audio recorded.
***************************************/

#include <pic18fregs.h>
#include "macros.h"
#include "spi.h"
#include "sd.h"
#include "fat.h"
#include "partitions.h"

code char at __CONFIG1H CONFIG1H = _OSC_INTIO7_1H & _FCMEN_OFF_1H & _IESO_OFF_1H;
code char at __CONFIG2L CONFIG2L = _PWRT_OFF_2L & _BOREN_OFF_2L & _WDT_DISABLED_CONTROLLED_2H;
code char at __CONFIG2H CONFIG2H = _WDT_DISABLED_CONTROLLED_2H;
code char at __CONFIG3H CONFIG3H = _LPT1OSC_OFF_3H & _MCLRE_MCLR_OFF_RE3_ON_3H ;
code char at __CONFIG4L CONFIG4L = _LVP_OFF_4L & _XINST_OFF_4L & _DEBUG_OFF_4L;
code char at __CONFIG5L CONFIG5L = ~0;
code char at __CONFIG5H CONFIG5H = _CPB_OFF_5H & _CPD_OFF_5H;
code char at __CONFIG6L CONFIG6L = 40;
code char at __CONFIG6H CONFIG6H = ~0;
code char at __CONFIG7L CONFIG7L = ~0;
code char at __CONFIG7H CONFIG7H = _EBTRB_OFF_7H;

//char num_to_7seg[10] = {0x3F,0x06,0x5B,0x4F,0x66,0x6D,0x7D,0x07,0x7F,0x67};
//char sndbuf[4][512];
char sndbuf1[512];
char sndbuf2[512];
char sndbuf3[512];
char sndbuf4[512];
int samplecount = 0;
volatile unsigned char swapped[4] = {1,1,1,1};
char filedesc = -1;
unsigned char buf_usage = 0;

struct fat_partition fat_part;

static void timer0_isr()  {
	if (buf_usage == 0) {
		sndbuf1[samplecount++] = (ADRESH-128);
	} else if (buf_usage == 1) {
		sndbuf2[samplecount++] = (ADRESH-128);
	} else if (buf_usage == 2) {
		sndbuf3[samplecount++] = (ADRESH-128);
	} else {
		sndbuf4[samplecount++] = (ADRESH-128);
	}

	/*unsigned short val = (ADRESH<<8)|ADRESL;
	val -= 32768;
	if (buf_usage == 0) {
		sndbuf1[samplecount++] = val&0xFF;
		sndbuf1[samplecount++] = (val>>8)&0xFF;
	} else if (buf_usage == 1) {
		sndbuf2[samplecount++] = val&0xFF;
		sndbuf2[samplecount++] = (val>>8)&0xFF;
	} else if (buf_usage == 2) {
		sndbuf3[samplecount++] = val&0xFF;
		sndbuf3[samplecount++] = (val>>8)&0xFF;
	} else {
		sndbuf4[samplecount++] = val&0xFF;
		sndbuf4[samplecount++] = (val>>8)&0xFF;
	}*/

	if (samplecount == 512) {
		samplecount = 0;
		swapped[buf_usage] = 0;
		buf_usage++;
		if (buf_usage == 4) buf_usage = 0;
	}
	
	ADCON0 |= 0x2; // GO!
	TMR0_CLR_FLAG();

	TMR0L = 131;
	EI();
}

unsigned long part_start = 0;
unsigned long read_func (void* desc, void * ptr, unsigned long s) {
	ReadSDBlock(ptr,s+part_start);
	return 512;
}

unsigned long write_func (void* desc, void * ptr, unsigned long s) {
	WriteSDBlock(ptr,s+part_start);
	return 512;
}

extern unsigned char wav_header[44];

void main() {
	short freq = 3;
	short work = 50;
	short i,j;
	char done = 1;
	TRISB = 0x00;
	PORTB = 0x00;
	TRISA = ~2;
	PORTA = 0;
	DI();

	// AD Setup
	ADCON1 = 0x0E;  // PORTA.b0 as ADC
	ADCON2 = 0x34;  // Fosc/64 and 16Tad
	ADRESH = 0;
	ADRESL = 0;
	ADCON0 = 1;  // Enabled!

	// Clock
	OSCCONbits.IRCF = 0x7;
	OSCTUNEbits.PLLEN = 1;
	for (i = 0; i < 10000; i++) {
		_asm
		nop
		_endasm;
	}

	// Timer
	TMR0L = 125;  // 256-125, we have to count 1000 betwen each sample (8Khz -> 8Khz)
	//T0CON = 0x02 | 0x40 | 0x80; // Enabled with prescaler to 1:8
	SET_TMR0(1,1,2) // 8KHz
	//SET_TMR0(1,1,1) // 16Khz
	TMR0_CLR_FLAG()
	EI_TMR0()
	
	file_system_init();
	InitSDCard();

/*	open_partition(&fat_part, 0, read_func, write_func);
	filedesc = file_open(&fat_part, "/F1", 0);
	PORTB = 0x01;
	for (i = 0; i < 30000; i++) {}
	for (i = 0; i < 30000; i++) {}
	PORTB = 0x0F;
	for (i = 0; i < 1024*2*1; i++) {
		file_write(&fat_part, filedesc, sndbuf1, 512);
	}
	PORTB = 0xF0;
	while(1);*/

	ReadSDBlock(sndbuf1,0);
	read_partition_table(0, sndbuf1, &part_start, 0);
	open_partition(&fat_part, 0, read_func, write_func);
	filedesc = file_open(&fat_part, "/F1", 0);
	file_write(&fat_part, filedesc, wav_header, sizeof(wav_header));
	for (i = 0; i < 256; i++) sndbuf1[i] = 0;
	file_write(&fat_part, filedesc, sndbuf1,512-sizeof(wav_header));
	
	PORTB = 0x3;

	EI();
	while(1) {
		while (swapped[0] == 1);
		file_write(&fat_part, filedesc, sndbuf1, 512);
		swapped[0] = 1;
		while (swapped[1] == 1);
		file_write(&fat_part, filedesc, sndbuf2, 512);
		swapped[1] = 1;
		while (swapped[2] == 1);
		file_write(&fat_part, filedesc, sndbuf3, 512);
		swapped[2] = 1;
		while (swapped[3] == 1);
		file_write(&fat_part, filedesc, sndbuf4, 512);
		swapped[3] = 1;
	}

	PORTB = 0x37;
	while(1);

}

static void isr() interrupt 1 {
	// Check the interruptions flags
	// from most likely to least
	if (TMR0_FLAG())
		timer0_isr();
}
