
/**************************************
 Copyright 2011 David Guillen Fandos
 Use at your own risk!
***************************************/

#include <pic18fregs.h>
#include "spi.h"
#include "macros.h"

// Sets up the SPI bus
void enable_spi_master(char capture_middle, char cke, char clk_high_idle, char speed) {
	// Init the FIFOs
	//in_fifo.pos = 0; in_fifo.len = 0;
	//out_fifo.pos = 0; out_fifo.len = 0;
	// Put the I/O Pins Tristates
	// RC5 is SDO, RC4 is SDI, RC3 is SCK
	TRISCbits.TRISC5 = TRISTATE_PIN_OUT;
	TRISCbits.TRISC4 = TRISTATE_PIN_IN;
	TRISCbits.TRISC3 = TRISTATE_PIN_OUT;
	// Set Up Config registers
	SSPCON1 = speed&0xF;
	SSPCON1bits.CKP = clk_high_idle;
	SSPSTAT = 0;
	//SSPSTAT = ((capture_middle&1)<<7) | ((cke&1)<<6);
	SSPCON1 |= 0x20;
	// Clear int flags + enable SPI interrupt
	SPI_CLR_FLAG();
	SPI_DI();
}

// Sends a buffer of data while receives another
void spi_sync_buf_sndrcv(char * send, char * recv, int num) {
	while (num-- > 0) {
		SPI_CLR_TX();
		SPI_PUT_BYTE(*send++);

		while (!SPI_TX());

		*recv++ = SPI_GET_BYTE();
		SPI_CLR_FLAG();
	}
}

// Sends a buffer of data and ignores the reception
void spi_sync_buf_snd(char * send, int num) {
	char read;
	while (num--) {
		SPI_CLR_TX();
		SPI_PUT_BYTE(*send++);

		while (!SPI_TX());
		SPI_CLR_FLAG();
	}
	read = SPI_GET_BYTE();
}

// Receives a buffer of data by sending dummy data
void spi_sync_buf_rcv(char * recv, int num) {
	while (num--) {
		SPI_CLR_TX();
		SPI_PUT_BYTE(0xFF);
		while (!SPI_TX());
		SPI_CLR_FLAG();
		*recv++ = SPI_GET_BYTE();
	}
}

// Send and recieve a byte sync
char spi_sync_send(char bdata) {
	char read;

	SPI_CLR_TX();
	SPI_PUT_BYTE(bdata);

	while (!SPI_TX()); // Wait the data to be sent

	read = SPI_GET_BYTE();

	SPI_CLR_FLAG();

	return read;
}

